/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim: set et sw=2 ts=2: */
/***************************************************************************
 *            macroheaderparser.cc
 *
 *  Wed Jul 22 08:42:23 CEST 2009
 *  Copyright 2009 Bent Bisballe Nyeng
 *  deva@aasimon.org
 ****************************************************************************/

/*
 *  This file is part of Pracro.
 *
 *  Pracro is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Pracro is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Pracro; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
#include "macroheaderparser.h"

#include <stdio.h>

// For assert
#include <assert.h>

// For open and friends
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// For vprintf and friends
#include <stdarg.h>

#include <errno.h>
#include <string.h>

#include "debug.h"
#include "configuration.h"
#include "exception.h"

void MacroHeaderParser::error(const char* fmt, ...)
{
  PRACRO_ERR_LOG(macro, "Error in MacroHeaderParser: ");

  {
    va_list argp;
    va_start(argp, fmt);
    PRACRO_ERR_LOG_VA(macro, fmt, argp);
    va_end(argp);

    fprintf(stderr, "\n");
  }

  {
    char *p;
    va_list argp;
    va_start(argp, fmt);
    if(vasprintf(&p, fmt, argp) != -1) {
      throw Exception("Error in MacroHeaderParser: " + std::string(p));
      free(p);
    }
    va_end(argp);
  }

}

MacroHeaderParser::MacroHeaderParser(std::string macrofile)
{
  m = NULL;

  file = macrofile;

  PRACRO_DEBUG(macro, "Using macro file: %s\n", macrofile.c_str());

  fd = open(macrofile.c_str(), O_RDONLY);
  if(fd == -1) error("Could not open file %s", macrofile.c_str());
}

MacroHeaderParser::~MacroHeaderParser()
{
  if(fd != -1) close(fd);
  if(m) delete m;
}

void MacroHeaderParser::startTag(std::string name, std::map< std::string, std::string> attributes)
{
  if(m) return;

  if(name == "macro") {
    m = new Macro();
    m->attributes = attributes;
  } else {
    throw Exception("Missing root tag 'macro' - found '" + name + "'");
  }
}

int MacroHeaderParser::readData(char *data, size_t size)
{
  if(m) return 0; // If m is allocated, it means that we have parsed the macro
                  // tag, and can dismiss the rest of the document.

  if(fd == -1) {
    PRACRO_ERR_LOG(macro, "Invalid file descriptor.\n");
    return 0;
  }
  ssize_t r = read(fd, data, size);
  if(r == -1) {
    PRACRO_ERR_LOG(macro, "Could not read...%s\n", strerror(errno));
    return 0;
  }
  return r;
}

void MacroHeaderParser::parseError(char *buf, size_t len, std::string error, int lineno)
{
  if(m) return; // Ignore "unclosed token" errors when the macro tag has been parsed.

  PRACRO_ERR_LOG(macro, "MacroHeaderParser[%s] error at line %d: %s\n", file.c_str(), lineno, error.c_str());
  PRACRO_ERR_LOG(macro, "\tBuffer %u bytes: [", len);
  if(fwrite(buf, len, 1, stderr) != len) {}
  PRACRO_ERR_LOG(macro, "]\n");

  char *slineno;
  if(asprintf(&slineno, " at line %d\n", lineno) != -1) {
    throw Exception(error + slineno);
    free(slineno);
  }
}

Macro *MacroHeaderParser::getMacro()
{
  return m;
}

#ifdef TEST_MACROHEADERPARSER

#define XMLFILE "/tmp/test_macroheaderparser.xml"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <memory.h>

static char xml[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<macro name=\"testmacro\" version=\"1.0\">\n"
" <sometag/>\n"
" <someothertag/>\n"
"</macro>"
;

static char xml_nonmacro[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<dims name=\"testmacro\" version=\"1.0\">\n"
" <sometag/>\n"
" <someothertag/>\n"
"</dims>"
;

static char xml_fail[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<macro name\"testmacro\" version=\"1.0\">\n"
" <sometag/>\n"
" <someothertag/>\n"
"</macro>"
;

int main()
{
  FILE *fp = fopen(XMLFILE, "w");
  if(!fp) {
    printf("Could not write to %s\n", XMLFILE);
    return 1;
  }
  fprintf(fp, xml);
  fclose(fp);

  {
    // Test parsing of correct macro xml data.
    MacroHeaderParser parser(XMLFILE);
    try {
      parser.parse();
    } catch(Exception &e) {
      printf("Failed to parse: %s\n", e.what());
      return 1;
    }
  }  

  fp = fopen(XMLFILE, "w");
  if(!fp) {
    printf("Could not write to %s\n", XMLFILE);
    return 1;
  }
  fprintf(fp, xml_nonmacro);
  fclose(fp);

  // Test parsing of correct xml data, but not macro (should throw an exception).
  {
    MacroHeaderParser parser(XMLFILE);
    try {
      parser.parse();
    } catch(Exception &e) {
      printf("Failed to parse: %s\n", e.what());
      goto onandon;
    }
    return 1;
  }
  onandon:

  fp = fopen(XMLFILE, "w");
  if(!fp) {
    printf("Could not write to %s\n", XMLFILE);
    return 1;
  }
  fprintf(fp, xml_fail);
  fclose(fp);

  // Test parsing of invalid xml data (should throw an exception).
  {
    MacroHeaderParser parser(XMLFILE);
    try {
      parser.parse();
    } catch(Exception &e) {
      printf("Failed to parse: %s\n", e.what());
      goto yetonandon;
    }
    return 1;
  }
  yetonandon:

  unlink(XMLFILE);

  return 0;
}

#endif/*TEST_MACROHEADERPARSER*/
