/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim: set et sw=2 ts=2: */
/***************************************************************************
 *            macrolist.cc
 *
 *  Wed Jul 22 10:26:40 CEST 2009
 *  Copyright 2009 Bent Bisballe Nyeng
 *  deva@aasimon.org
 ****************************************************************************/

/*
 *  This file is part of Pracro.
 *
 *  Pracro is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Pracro is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Pracro; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
#include "macrolist.h"

#include <sys/types.h>
#include <dirent.h>

#include "debug.h"
#include "macroheaderparser.h"

static std::vector<std::string> listdir(std::string path)
{
  std::vector<std::string> files;

  DIR* dir = opendir(path.c_str());
  if(!dir) {
    PRACRO_ERR(dump, "Could not open directory: %s\n", path.c_str());
    return files;
  }

  struct dirent *d;
  while((d = readdir(dir)) != 0) {
    //if(d->d_type == DT_DIR) {
    std::string name = d->d_name;
    if(name.length() >= 4 && name.substr(name.length() - 4) == ".xml")
      files.push_back(name);
    //}
  }
  closedir(dir);

  return files;
}

MacroList::MacroList(std::string macropath)
{
  this->macropath = macropath;
  std::vector<std::string> macros = listdir(macropath);
  std::vector<std::string>::iterator i = macros.begin();
  while(i != macros.end()) {
    MacroHeaderParser parser(macropath + "/" + *i);
    try {
      parser.parse();
      Macro *macro = parser.getMacro();
      (*this)[macro->attributes["name"]][VersionStr(macro->attributes["version"])] = *i;
    } catch(Exception &e) {
      PRACRO_WARN(macrolist, "Skipping %s: %s\n", i->c_str(), e.what());
    }
    i++;
  }

  {
    iterator i = begin();
    while(i != end()) {
      MacroListItem::iterator j = i->second.begin();
      while(j != i->second.end()) {
        PRACRO_DEBUG(macrolist, "%s - v%s file: %s\n",
                     i->first.c_str(),
                     ((std::string)j->first).c_str(),
                     j->second.c_str());
        j++;
      }
      i++;
    }
  }
}

std::string MacroList::getLatestVersion(std::string macro) throw(Exception)
{
  if(find(macro) == end()) throw Exception("Macro ["+macro+"] does not exist");
  MacroListItem mli = (*this)[macro];
  if(mli.size() == 0) return "";
  PRACRO_DEBUG(macrolist, "Search for %s - found %s v%s\n",
               macro.c_str(),
               (macropath + "/" + mli.begin()->second).c_str(),
               ((std::string)mli.begin()->first).c_str());
  return macropath + "/" + mli.begin()->second;
}

#ifdef TEST_MACROLIST

#define MACRODIR "/home"  // We assume this directory exists and does not contain any xml files!

int main()
{
  // Test sorting
  MacroList lst(MACRODIR);

  lst["macro1"][VersionStr("1.0")] = "macro1-1.0.xml";
  lst["macro1"][VersionStr("1.1")] = "macro1-1.1.xml";
  lst["macro1"][VersionStr("1.1.1")] = "macro1-1.1.1.xml";
  lst["macro1"][VersionStr("1.2")] = "macro1-1.2.xml";
  lst["macro2"][VersionStr("1.0")] = "macro2.xml";
  lst["macro3"][VersionStr("1.0")] = "macro3.xml";

  std::vector<std::string> refs;
  refs.push_back("macro1-1.2.xml");
  refs.push_back("macro1-1.1.1.xml");
  refs.push_back("macro1-1.1.xml");
  refs.push_back("macro1-1.0.xml");
  refs.push_back("macro2.xml");
  refs.push_back("macro3.xml");

  MacroList::iterator i = lst.begin();
  std::vector<std::string>::iterator k = refs.begin();
  while(i != lst.end()) {
    MacroListItem::iterator j = i->second.begin();
    while(j != i->second.end()) {
      printf("%s - v%s file: %s - should be %s\n",
             i->first.c_str(),
             ((std::string)j->first).c_str(),
             j->second.c_str(),
             k->c_str());
      if(j->second != *k) return 1;
      j++; k++;
    }
    i++;
  }

  // Test lookup of latest versions.
  std::string m1 = lst.getLatestVersion("macro1");
  printf("Latest macro1: %s\n", m1.c_str());
  if(m1 != MACRODIR"/macro1-1.2.xml") return 1;

  std::string m2 = lst.getLatestVersion("macro2");
  printf("Latest macro2: %s\n", m2.c_str());
  if(m2 != MACRODIR"/macro2.xml") return 1;

  std::string m3 = lst.getLatestVersion("macro3");
  printf("Latest macro3: %s\n", m3.c_str());
  if(m3 != MACRODIR"/macro3.xml") return 1;

  // Look for non existing macro (this should throw an exception)
  try {
    std::string m4 = lst.getLatestVersion("macro4");
  } catch(Exception &e) {
    printf("ERROR: %s\n", e.what());
    goto onandon;
  }
  return 1;
 onandon:

  return 0;
}

#endif/*TEST_MACROLIST*/
