/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/***************************************************************************
 *            server.cc
 *
 *  Wed Aug 22 12:16:03 CEST 2007
 *  Copyright 2007 Bent Bisballe Nyeng
 *  deva@aasimon.org
 ****************************************************************************/

/*
 *  This file is part of Pracro.
 *
 *  Pracro is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Pracro is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Pracro; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
#include "server.h"

#include "tcpsocket.h"
#include <errno.h>

#include <stdlib.h>

// For fork
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include "configuration.h"
#include "transaction.h"
#include "transactionparser.h"
#include "templateparser.h"
#include "macroparser.h"

#include "queryhandler.h"
#include "queryhandlerpracro.h"
#include "queryhandlerpentominos.h"

#include "queryparser.h"
#include "luaquerymapper.h"
#include "database.h"
#include "widgetgenerator.h"
#include "resumeparser.h"
#include "journal_commit.h"
#include "xml_encode_decode.h"

#include "macrolist.h"
#include "templatelist.h"
#include "versionstr.h"

static std::string error_box(std::string message)
{
  std::string errorbox =
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
    "<pracro version=\"1.0\">\n"
    "  <template name=\"error\">\n"
    "    <macro name=\"error\" static=\"true\">\n"
    "      <widgets caption=\"ERROR!\" layout=\"vbox\" name=\"error\">\n"
    "        <textedit name=\"errorlabel\" value=\"" + message + "\"/>\n"
    "      </widgets>\n"
    "    </macro>\n"
    "  </template>\n"
    "</pracro>\n";
  return errorbox;
}

class NotFoundException : public Exception {
public:
  NotFoundException(Request &r)
  : Exception("Macro " + r.macro + " not found in template " + r.templ) {}
};


static std::string handleCommits(Transaction *transaction, Database &db, 
                                 JournalWriter &journalwriter, MacroList &macrolist,
                                 TemplateList &templatelist)
{
  std::string answer;

  Commits::iterator i = transaction->commits.begin();
  while(i != transaction->commits.end()) {
    Commit &commit = *i;
    
    MacroParser mp(macrolist.getLatestVersion(commit.macro));
    mp.parse();
    Macro *macro = mp.getMacro();
    
    std::string resume = resume_parser(macro->resume, commit);
    commit.fields["journal.resume"] = resume;
    db.commitTransaction(transaction->user, transaction->cpr, *macro, commit.fields);
    
    if(resume != "") {

      TemplateParser tp(templatelist.getLatestVersion(commit.templ));
      tp.parse();
      Template *templ = tp.getTemplate();

      journalwriter.addEntry(*transaction, commit, resume, templ);
    }
    
    i++;
  }

  return answer;
}


static std::string handleRequest(Transaction *transaction,
                                 TCPSocket &pentominos_socket,
                                 Database &db,
                                 JournalWriter &journalwriter,
                                 MacroList &macrolist,
                                 TemplateList &templatelist)
{
  std::string answer;

  Requests::iterator i = transaction->requests.begin();
  while(i != transaction->requests.end()) {
    Request &request = *i;
    
    PRACRO_DEBUG(server, "Handling request - macro: %s, template: %s\n",
                 request.macro.c_str(), request.templ.c_str());
    
    // Read and parse the template file.
    TemplateParser tp(templatelist.getLatestVersion(request.templ));
    tp.parse();
    
    Template *templ = tp.getTemplate();
    
    answer += "  <template name=\"";
    answer += templ->attributes["name"];
    answer += "\" title=\"";
    answer += templ->attributes["title"];
    answer += "\">\n";
    
    bool foundmacro = false;
    
    // Generate the macro and return it to the client
    std::vector< Macro >::iterator mi2 = templ->macros.begin();
    while(mi2 != templ->macros.end()) {
      Macro &macro = (*mi2);

      if(macro.isHeader) {
        answer += "    <header caption=\"" + macro.attributes["caption"] + "\"/>\n";
        mi2++;
        continue;
      }

      bool completed = db.checkMacro(transaction->cpr,
                                     macro.attributes["name"],
                                     time(NULL)-Conf::db_max_ttl);

      answer += "    <macro completed=";
      if(completed) answer += "\"true\"";
      else answer += "\"false\"";

      std::map< std::string, std::string >::iterator ai = macro.attributes.begin();
      while(ai != macro.attributes.end()) {
        std::string name = ai->first;
        std::string value = ai->second;
        answer += " "+name+"=\"" + value + "\"";
        ai++;
      }

      if(macro.attributes["name"] == request.macro || 
         (macro.attributes.find("static") != macro.attributes.end() &&
          macro.attributes["static"] == "true")
         ) {
        foundmacro = true;

        MacroParser mp(macrolist.getLatestVersion(macro.attributes["name"]));
        mp.parse();
        Macro *m = mp.getMacro();
        answer += " caption=\"" + m->widgets.attributes["caption"] + "\"";
        answer += ">\n";

        LUAQueryMapper lqm;

        ////////////////////////
        std::vector< Query >::iterator qi = m->queries.begin();
        while(qi != m->queries.end()) {

          Query &query = *qi;
          std::string service = query.attributes["service"];

          if(service == "pentominos") {
            // Send the queries to Pentominos (if any)
            QueryHandlerPentominos qh(pentominos_socket, transaction->cpr);

            QueryResult queryresult = qh.exec(*qi);
            lqm.addQueryResult(queryresult);
          }

          if(service == "pracro") {
            // Send the queries to Pentominos (if any)
            QueryHandlerPracro qh(db, transaction->cpr);

            QueryResult queryresult = qh.exec(*qi);
            lqm.addQueryResult(queryresult);
          }

          qi++;
        }

        // Handle scripts
        if(m->scripts.size()) {
          answer += "      <scripts>\n";
            
          std::vector< Script >::iterator spi = m->scripts.begin();
          while(spi != m->scripts.end()) {
            answer += "        <script language=\"" + spi->attributes["language"] 
              + "\" name=\"" + spi->attributes["name"] + "\">\n";
            answer += xml_encode(spi->attributes["code"]);
            answer += "\n        </script>\n";
            spi++;
          }
          answer += "      </scripts>\n";
        }

        answer += widgetgenerator(transaction->cpr, *m, lqm, db);
      } else {
        // only find macro title
        MacroParser mp(macrolist.getLatestVersion(macro.attributes["name"]));
        mp.parse();
        Macro *m = mp.getMacro();
        answer += " caption=\"" + m->widgets.attributes["caption"] + "\"";
        answer += ">\n";

      }

      if(completed) {
        answer += "      <resume>";
        answer += db.getResume(transaction->cpr, macro, time(NULL) - Conf::db_max_ttl);
        answer += "</resume>\n";
      }

      answer += "    </macro>\n";
      mi2++;

    }
      
    if(foundmacro == false && request.macro != "")
      throw NotFoundException(request);

    answer += "  </template>\n";

    i++;
  }

  return answer;
}

static std::string handleTransaction(Transaction *transaction,
                                     TCPSocket &pentominos_socket,
                                     Database &db,
                                     JournalWriter &journalwriter,
                                     MacroList &macrolist,
                                     TemplateList &templatelist)
{
  std::string answer;
  answer += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
  answer += "<pracro version=\"1.0\">\n";

  try {
    answer += handleCommits(transaction, db, journalwriter, macrolist, templatelist);
  } catch( std::exception &e ) {
    PRACRO_ERR(server, "Commit error: %s\n", e.what());
    return error_box(xml_encode(e.what()));
  }

  try {
    answer += handleRequest(transaction, pentominos_socket, db, journalwriter,
                            macrolist, templatelist);
  } catch( std::exception &e ) {
    PRACRO_ERR(server, "Request error: %s\n", e.what());
    return error_box(xml_encode(e.what()));
  }

  answer += "</pracro>\n";

  PRACRO_DEBUG(server, "Done handling transaction\n");
  PRACRO_DEBUG(serverxml, "%s\n", answer.c_str());
  return answer;
}


static void handleConnection(TCPSocket *socket)
{
  TCPSocket pentominos_socket;
#ifndef WITHOUT_PENTOMINOS
  pentominos_socket.connect(Conf::pentominos_addr, Conf::pentominos_port);
#endif/*WITHOUT_PENTOMINOS*/

  Database db(Conf::database_backend, Conf::database_addr, "", Conf::database_user, Conf::database_passwd, "");

  JournalWriter journalwriter(Conf::journal_commit_addr.c_str(), Conf::journal_commit_port);

  MacroList macrolist(Conf::xml_basedir + "/macros");
  TemplateList templatelist(Conf::xml_basedir + "/templates");

  ssize_t size;
  char buf[4096];
  
  Transaction *transaction = NULL;
  TransactionParser *parser = NULL;
  
  //  while( (size = socket->read(buf, sizeof(buf))) != -1) {
  while( (size = socket->read(buf, sizeof(buf))) > 0) {

    PRACRO_DEBUG(server, "Read %d bytes from network\n", size);
   
    while(size) {

      if(transaction == NULL) {
        transaction = new Transaction();
      }

      if(parser == NULL) {
        parser = new TransactionParser(transaction);
      }

      PRACRO_DEBUG(server, "Got %d bytes in read loop\n", size);
      if(parser->parse(buf, size)) {
        PRACRO_DEBUG(server, "Got complete XML document %d bytes used, %d bytes in current buffer.\n", parser->usedBytes(), size);

        socket->write(handleTransaction(transaction, pentominos_socket,
                                        db, journalwriter, macrolist, templatelist));
        size = size - parser->usedBytes();
        if(size) {
          strcpy(buf, buf + parser->usedBytes());
          PRACRO_DEBUG(server, "Replaying %d bytes.\n", size);
        }
        
        delete transaction; transaction = NULL;
        delete parser; parser = NULL;
      } else {
        size = 0;
        memset(buf, 0, sizeof(buf));
      }
    }
  }

  if(transaction) {
    delete transaction;
    transaction = NULL;
  }

  if(parser) {
    delete parser;
    parser = NULL;
  }

  journalwriter.commit();

  PRACRO_DEBUG(server, "Out of read loop!\n");
}

//#define NON_FORKING
#include <sys/socket.h>
extern bool pracro_is_running;
void server()
{
  port_t port = Conf::server_port;
  TCPSocket *socket = NULL;
  
  try {
    socket = new TCPSocket("Listen socket");
    socket->listen(port);
  } catch (Exception &e) {
    PRACRO_ERR_LOG(server, "Error in listen:\n%s\n", e.what());
    delete socket;
    socket = NULL;
    return;
  }

  while(pracro_is_running && socket->connected()) {

    { // Reload if new port is assigned.
      int old_port = port;
      port = Conf::server_port;

      if(port != old_port) {
        // Start listening on the new port
        delete socket;
        socket = new TCPSocket("Listen socket (reloaded)");
        socket->listen(port);
      }
    }

    TCPSocket *child = socket->accept();
    if(child) {

#ifndef NON_FORKING
      switch(fork()) {
      case -1: // error
        PRACRO_ERR_LOG(server, "Could not fork: %s\n", strerror(errno));
        break;
        
      case 0: // child
        delete socket;
#endif/*NON_FORKING*/
        handleConnection(child);
        delete child;
#ifndef NON_FORKING
        return;
        
      default: // parent
        delete child;
        break;
      }
#endif/*NON_FORKING*/

    }
  }

  //socket->shutdown();
  delete socket;

  PRACRO_DEBUG(server, "Server gracefully shut down.\n");
}


#ifdef TEST_SERVER

#include <sys/types.h>
#include <signal.h>

bool pracro_is_running = true;

char request[] = 
  "<?xml version='1.0' encoding='UTF-8'?>\n"
  "<pracro cpr=\"2003791613\" version=\"1.0\">\n"
  "  <request macro=\"example\" template=\"example\"/>\n"
  "</pracro>\n";

int main()
{
  Conf::xml_basedir = "../xml/";
  Conf::server_port = 32100; // Make sure wo don't interrupt an already running server.
  Conf::database_backend = "testdb";
  pid_t pid = fork();
  
  switch(pid) {
  case -1: // error
    perror("fork() failed!\n");
    return 1;
    
  case 0: // child
    try {
      server();
    } catch(Exception &e) {
      printf(e.what());
      return 1;
    }
    return 0;
    
  default: // parent
    try {
      //      sleep(1); // Make sure the server is started.
      TCPSocket socket;
      socket.connect("localhost", Conf::server_port);
      
      socket.write(request);

      //      sleep(1); // Make sure the server has handled the request.
      char buf[32];
      memset(buf, 0, sizeof(buf));
      //      while(socket.read(buf, 31, 1000)) {
      while(socket.read(buf, 31, 1000000)) {
        printf(buf); fflush(stdout);
        memset(buf, 0, sizeof(buf));
      }
    } catch(Exception &e) {
      printf(e.what());
      kill(pid, SIGKILL); // Kill the server again.
      return 1;
    }
    kill(pid, SIGKILL); // Kill the server again.
    return 0;
  }
  
  return 1;
}

#endif/*TEST_SERVER*/
