/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/***************************************************************************
 *            templateparser.cc
 *
 *  Mon May 12 08:36:24 CEST 2008
 *  Copyright 2008 Bent Bisballe Nyeng
 *  deva@aasimon.org
 ****************************************************************************/

/*
 *  This file is part of Pracro.
 *
 *  Pracro is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Pracro is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Pracro; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
#include "templateparser.h"
#include "configuration.h"
#include "debug.h"

// For assert
#include <assert.h>

// For open and friends
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// For vprintf and friends
#include <stdarg.h>

#include <errno.h>
#include <string.h>

#include "exception.h"

void TemplateParser::error(const char* fmt, ...)
{
  va_list argp;
  char *p;
  va_start(argp, fmt);
  vasprintf(&p, fmt, argp);
  va_end(argp);
  PRACRO_ERR_LOG(template, "Error in TemplateParser: %s\n", p);
  throw Exception(std::string("Error in TemplateParser: ") + p);
  free(p);
}

TemplateParser::TemplateParser(std::string templatefile)
{
  state = UNDEFINED;
  t = new Template();
  current_macro = NULL;

  file = templatefile;

  PRACRO_DEBUG(template, "Using template file: %s\n", file.c_str());

  fd = open(file.c_str(), O_RDONLY);
  if(fd == -1) error("Could not open file %s", file.c_str());
}

TemplateParser::~TemplateParser()
{
  if(fd != -1) close(fd);
  delete t;
}

void TemplateParser::characterData(std::string &data)
{
}

void TemplateParser::startTag(std::string name, std::map< std::string, std::string> attributes)
{
  // Enable macro parsing
  if(name == "template") {
    if(state != UNDEFINED) error("Template found not a root node.");
    state = TEMPLATE;

    assert(t); // A Template has not yet been allocated, cannot create template!

    t->attributes = attributes;

    return;
  }

  // Create macro and enable parsing of queries, maps and widgets
  if(name == "macro" || name == "header") {
    if(state != TEMPLATE) error("macro found outside template.");
    state = MACRO;

    assert(t); // A Template has not yet been allocated, cannot create macro!
    
    Macro m;
    m.attributes = attributes;
    m.isHeader = name == "header";
    t->macros.push_back(m);
    current_macro = &(t->macros.back());

    return;
  }

  error("Unknown/illegal tag: %s", name.c_str());
}

void TemplateParser::endTag(std::string name)
{
  if(name == "template") state = UNDEFINED;
  if(name == "macro" || name == "header") {
    current_macro = NULL;
    state = TEMPLATE;
  }
}

int TemplateParser::readData(char *data, size_t size)
{
  if(fd == -1) {
    PRACRO_ERR_LOG(template, "Invalid file descriptor.\n");
    return 0;
  }
  ssize_t r = read(fd, data, size);
  if(r == -1) {
    PRACRO_ERR_LOG(template, "Could not read...%s\n", strerror(errno));
    return 0;
  }
  return r;
}

void TemplateParser::parseError(char *buf, size_t len, std::string error, int lineno)
{
  fprintf(stderr, "TemplateParser[%s] error at line %d: %s\n", file.c_str(), lineno, error.c_str());
  fprintf(stderr, "\tBuffer %u bytes: [", len);
  if(fwrite(buf, len, 1, stderr) != len) {}
  fprintf(stderr, "]\n");
  fflush(stderr);

  char *slineno;
  if(asprintf(&slineno, " at line %d\n", lineno) != -1) {
    throw Exception(error + slineno);
    free(slineno);
  }
}

Template *TemplateParser::getTemplate()
{
  return t;
}

#ifdef TEST_TEMPLATEPARSER

#define XMLFILE "/tmp/test_templateparser.xml"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <memory.h>

static char xml[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<template name=\"testtemplate\" version=\"1.0\">\n"
" <macro name=\"mymacro\"/>\n"
" <header caption=\"mycaption\"/>\n"
"</template>"
;

static char xml_nontemplate[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<dims name=\"testtemplate\" version=\"1.0\">\n"
" <sometag/>\n"
" <someothertag/>\n"
"</dims>"
;

static char xml_fail[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<template name\"testtemplate\" version=\"1.0\">\n"
" <sometag/>\n"
" <someothertag/>\n"
"</template>"
;

int main()
{
  FILE *fp = fopen(XMLFILE, "w");
  if(!fp) {
    printf("Could not write to %s\n", XMLFILE);
    return 1;
  }
  fprintf(fp, xml);
  fclose(fp);

  {
    // Test parsing of correct template xml data.
    TemplateParser parser(XMLFILE);
    try {
      parser.parse();
    } catch(Exception &e) {
      printf("Failed to parse: %s\n", e.what());
      return 1;
    }
  }  

  fp = fopen(XMLFILE, "w");
  if(!fp) {
    printf("Could not write to %s\n", XMLFILE);
    return 1;
  }
  fprintf(fp, xml_nontemplate);
  fclose(fp);

  // Test parsing of correct xml data, but not template (should throw an exception).
  {
    TemplateParser parser(XMLFILE);
    try {
      parser.parse();
    } catch(Exception &e) {
      printf("Failed to parse: %s\n", e.what());
      goto onandon;
    }
    return 1;
  }
  onandon:

  fp = fopen(XMLFILE, "w");
  if(!fp) {
    printf("Could not write to %s\n", XMLFILE);
    return 1;
  }
  fprintf(fp, xml_fail);
  fclose(fp);

  // Test parsing of invalid xml data (should throw an exception).
  {
    TemplateParser parser(XMLFILE);
    try {
      parser.parse();
    } catch(Exception &e) {
      printf("Failed to parse: %s\n", e.what());
      goto yetonandon;
    }
    return 1;
  }
  yetonandon:

  unlink(XMLFILE);

  return 0;
}

#endif/*TEST_TEMPLATEPARSER*/
