/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/***************************************************************************
 *            transactionparser.cc
 *
 *  Fri Aug 31 09:30:06 CEST 2007
 *  Copyright 2007 Bent Bisballe Nyeng
 *  deva@aasimon.org
 ****************************************************************************/

/*
 *  This file is part of Pracro.
 *
 *  Pracro is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Pracro is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Pracro; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
#include "transactionparser.h"

#include <stdio.h>
#include <string.h>
#include <expat.h>

#include <string>
#include <map>

#include "debug.h"
#include "exception.h"

static void error(const char* fmt, ...)
{
  PRACRO_ERR_LOG(transactionparser, "Error in TransactionParser: ");

  {
    va_list argp;
    va_start(argp, fmt);
    PRACRO_ERR_LOG_VA(macro, fmt, argp);
    va_end(argp);

    fprintf(stderr, "\n");
  }

  {
    char *p;
    va_list argp;
    va_start(argp, fmt);
    if(vasprintf(&p, fmt, argp) != -1) {
      throw Exception("Error in TransactionParser: " + std::string(p));
      free(p);
    }
    va_end(argp);
  }
}

TransactionParser::TransactionParser(Transaction *transaction)
{
  this->transaction = transaction;
  done = false;
  totalbytes = 0;
}

void TransactionParser::startTag(std::string name, std::map< std::string, std::string> attributes)
{
  if(name == "pracro") {
    transaction->user = attributes["user"];
    transaction->cpr = attributes["cpr"];
    transaction->version = attributes["version"];
  }

  if(name == "request") {
    Request r;
    r.templ = attributes["template"];
    r.macro = attributes["macro"];
    transaction->requests.push_back(r);
  }

  if(name == "commit") {
    Commit c;
    c.templ = attributes["template"];
    c.macro = attributes["macro"];
    c.version = attributes["version"];
    transaction->commits.push_back(c);
  }

  if(name == "field") {
    if(!transaction->commits.size()) error("Field without a commit tag!");
    if(attributes.find("name") == attributes.end()) error("Field is missing 'name' attribute");
    if(attributes.find("value") == attributes.end()) error("Field is missing 'value' attribute");
    transaction->commits.back().fields[attributes["name"]] = attributes["value"];
  }
}

void TransactionParser::parseError(char *buf, size_t len, std::string error, int lineno)
{
  PRACRO_ERR_LOG(transactionparser, "TransactionParser error at line %d: %s\n",
                 lineno, error.c_str());
  PRACRO_ERR_LOG(transactionparser, "\tBuffer %u bytes: [", len);
  if(fwrite(buf, len, 1, stderr) != len) {}
  PRACRO_ERR_LOG(transactionparser, "]\n");

  char *slineno;
  if(asprintf(&slineno, " at line %d\n", lineno) != -1) {
    throw Exception(error + slineno);
    free(slineno);
  }
}

#ifdef TEST_TRANSACTIONPARSER

static char xml_minimal[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<pracro/>\n"
  ;

static char xml_request[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<pracro version=\"1.0\" user=\"testuser\" cpr=\"1505050505\">\n"
" <request macro=\"test\" template=\"test\"/>\n"
"</pracro>\n"
  ;

static char xml_commit[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<pracro version=\"1.0\" user=\"testuser\" cpr=\"1505050505\">\n"
" <commit version=\"\" macro=\"referral\" template=\"amd_forunders\" >\n"
"  <field value=\"Some docs\" name=\"referral.doctor\"/>\n"
"  <field value=\"DIMS\" name=\"referral.diagnosecode\"/>\n"
"  <field value=\"Avs\" name=\"referral.diagnose\"/>\n"
" </commit>\n"
"</pracro>\n"
  ;

static char xml_fail[] =
"<?xml version='1.0' encoding='UTF-8'?>\n"
"<pracro version=\"1.0\" user=\"testuser\" cpr=\"1505050505\">\n"
" <request macro=\"test template=\"test\"/>\n"
"</pracro>\n"
  ;

int main()
{
  // Test minimal
  try {
    Transaction transaction;
    TransactionParser parser(&transaction);
    parser.parse(xml_minimal, strlen(xml_minimal));
  } catch(Exception &e) {
    printf("ERROR: %s\n", e.what());
    return 1;
  }

  // Test request
  try {
    Transaction transaction;
    TransactionParser parser(&transaction);
    parser.parse(xml_request, strlen(xml_request));
  } catch(Exception &e) {
    printf("ERROR: %s\n", e.what());
    return 1;
  }

  // Test commit
  try {
    Transaction transaction;
    TransactionParser parser(&transaction);
    parser.parse(xml_commit, strlen(xml_commit));
  } catch(Exception &e) {
    printf("ERROR: %s\n", e.what());
    return 1;
  }

  // Test parse error (should throw an exception)
  try {
    Transaction transaction;
    TransactionParser parser(&transaction);
    parser.parse(xml_fail, strlen(xml_fail));
  } catch(Exception &e) {
    printf("ERROR: %s\n", e.what());
    goto onandon;
  }
  printf("We should fail here...\n");
  return 1;

 onandon:

  return 0;
}

#endif/*TEST_TRANSACTIONPARSER*/
